import logging
import os

import dendropy

from metatree.exception import MetaTreeExit


class Batchfile(object):

    def __init__(self, path):
        self.logger = logging.getLogger('timestamp')
        self.path = path
        self.ref, self.data = self.read()

    def read(self):
        if not os.path.isfile(self.path):
            raise MetaTreeExit(f'The batchfile does not exist: {self.path}')
        out = dict()
        ref = None
        invalid_paths = list()
        with open(self.path) as fh:
            for line in fh.readlines():
                line = line.strip()
                if not line.startswith('#'):
                    tree_id, tree_path = line.split('\t')
                    out[tree_id] = tree_path
                    if ref is None:
                        ref = tree_id
                    if not os.path.isfile(tree_path):
                        invalid_paths.append((tree_id, tree_path))
        for tree_id, tree_path in invalid_paths:
            self.logger.error(f'The path for {tree_id} does not exist: {tree_path}')
        if len(invalid_paths) > 0:
            raise MetaTreeExit('Invalid tree paths were present in the batchfile.')
        return ref, out

    def common_taxa(self):
        out = set()
        for tree_id, tree_path in self.data.items():
            tree = dendropy.Tree.get_from_path(tree_path, schema='newick', preserve_underscores=True)
            cur_set = {x.label for x in tree.taxon_namespace}
            if len(out) == 0:
                out = cur_set
            else:
                out = out.intersection(cur_set)
        return out



import logging
import os

from metatree.exception import MetaTreeExit


class TaxonomyFile(object):

    def __init__(self, path):
        self.path = path
        self.data = self.read()
        self.logger = logging.getLogger('timestamp')

    def read(self):
        if not os.path.isfile(self.path):
            raise MetaTreeExit(f'The taxonomy file does not exist: {self.path}')
        out = dict()
        invalid_tax = list()
        with open(self.path) as fh:
            for line in fh.readlines():
                gid, tax = line.strip().split('\t')
                out[gid] = tax
                if len(tax.split(';')) != 7:
                    invalid_tax.append((gid, tax))
        for gid, tax in invalid_tax:
            self.logger.error(f'Genome {gid} does not have a valid taxonomy: {tax}')
        if len(invalid_tax) > 0:
            raise MetaTreeExit(f'There were {len(invalid_tax)} invalid taxonomies detected.')
        return out



import logging
import os

from metatree.common import make_sure_path_exists
from metatree.f_measure_tree import FMeasureTree
from metatree.io import Batchfile, RfResults
from metatree.io.taxonomy_file import TaxonomyFile
from metatree.tree_decorate import TreeDecorate
from metatree.tree_dist import TreeDist
from metatree.tree_root import TreeRoot


def run_pipeline(batchfile: Batchfile, out_dir: str, tax_file: TaxonomyFile, outgroup: str, cpus: int):
    logger = logging.getLogger('timestamp')

    # Setup output paths.
    dir_root = os.path.join(out_dir, 'intermediate_results', 'trees_rooted')
    dir_dec = os.path.join(out_dir, 'intermediate_results', 'trees_decorated')

    dir_rf_common = os.path.join(out_dir, 'results', 'robinson_foulds_common_taxa')
    dir_rf_all = os.path.join(out_dir, 'results', 'robinson_foulds_all_taxa')
    make_sure_path_exists(dir_rf_common)
    make_sure_path_exists(dir_rf_all)
    rf_common = RfResults(os.path.join(dir_rf_common, 'rf_common_taxa.tsv'))
    rf_all = RfResults(os.path.join(dir_rf_all, 'rf_all_taxa.tsv'))

    # tbl_diff = os.path.join(out_dir, 'results', 'model_taxonomy_diff.tsv')

    # Root the trees.
    tree_root = TreeRoot(dir_root)
    tree_root.run(batchfile, dir_root, outgroup, tax_file, cpus)

    # Decorate the trees.
    tree_decorate = TreeDecorate(dir_dec)
    tree_decorate.run(batchfile, dir_root, dir_dec, tax_file, cpus)

    # Pairwise comparison of all trees.
    td = TreeDist()
    td.run(rf_common, batchfile, dir_root, dir_dec, cpus, common_taxa=True)
    td.run(rf_all, batchfile, dir_root, dir_dec, cpus, common_taxa=False)

    # Summarise the pairwise distances (trees + heatmap)
    logger.info(f'Writing pairwise Robinson-Foulds distances for common taxa to: {dir_rf_common}')
    td.summarise_dist(rf_common, dir_rf_common)
    logger.info(f'Writing pairwise Robinson-Foulds distances for all taxa to: {dir_rf_all}')
    td.summarise_dist(rf_all, dir_rf_all)

    # Summarise the differences between all models and the reference.
    # mmt = MismatchTable(tbl_diff)
    # mmt.run_and_save(batchfile, dir_dec, tax_file)

    # Create the tree-of-trees comparison.
    fmt = FMeasureTree(tax_file.path)
    for tree_id, tree_path in batchfile.data.items():
        if tree_id != batchfile.ref:
            fmt.add_table(tree_id, os.path.join(dir_dec, f'{tree_id}_rooted_decorated.tree-table'))
    for legend in (True, False):
        if legend:
            the_path = os.path.join(out_dir, 'results', 'tree_comparison_legend.svg')
        else:
            the_path = os.path.join(out_dir, 'results', 'tree_comparison.svg')
        fmt.run(legend=legend, out_path=the_path)

    return



import os
from collections import defaultdict

import dendropy
import ete3

class FMeasureTree(object):
    """Compares Phylorank output tables and plots differences in a tree."""

    def __init__(self, path_taxonomy):
        """Initialises the class, requires the shared taxonomy file."""
        self.files = dict()
        self.tf = TaxonomyFile(path_taxonomy)

    def add_table(self, label, path):
        """Adds a table for comparison."""
        self.files[label] = FMeasureTable(path)

    def get_n_common(self):
        """Determine the number of genomes which are common between ALL models."""
        out = defaultdict(lambda: defaultdict(lambda: 0))

        # Index by rank.
        d_rank_model = defaultdict(dict)
        for model_id, fm in self.files.items():
            for rank, rank_dict in fm.get_content().items():
                d_rank_model[rank][model_id] = rank_dict

        # Iterate over each rank and find the number of common taxa.
        for rank, model_dict in d_rank_model.items():

            # This is common for all ranks.
            if set(self.files.keys()) == set(model_dict.keys()):

                # Determine the number of common taxa for all of the models.
                all_in, all_out = None, None
                for cur_model_id, cur_model_info in model_dict.items():
                    if all_in is None:
                        all_in = set(cur_model_info['rogue_in'])
                    else:
                        all_in = all_in.intersection(set(cur_model_info['rogue_in']))
                    if all_out is None:
                        all_out = set(cur_model_info['rogue_out'])
                    else:
                        all_out = all_out.intersection(set(cur_model_info['rogue_out']))

                out[rank]['in'] = len(all_in)
                out[rank]['out'] = len(all_out)
        return out

    def get_poly_ranks(self):
        out = set()
        for model_id, fm in self.files.items():
            for rank, rank_dict in fm.get_content().items():
                if rank_dict['f_measure'] < 1.0:
                    out.add(rank)
        return out

    @staticmethod
    def get_text_face(text, bg, opacity, fsize=8, bold=False):
        margin_size = 2

        tf = ete3.faces.TextFace(str(text), fsize=fsize, tight_text=True)
        tf.background.color = bg
        tf.margin_top = margin_size
        tf.margin_bottom = margin_size
        tf.margin_right = margin_size
        tf.margin_left = margin_size
        tf.opacity = opacity
        tf.rotation = -0
        tf.bold = bold
        tf.border.color = '#f2f2f2'
        tf.border.width = None  # try 1?
        tf.hz_align = 1
        tf.vt_align = 1
        return tf

    def run(self, legend, out_path, rotation_deg=0):

        # Required for the script to be able to render an output.
        os.environ['QT_QPA_PLATFORM'] = 'offscreen'

        # Create a newick tree spanning all nodes identified in f-measure tables.
        newick = NewickTree(self.tf)
        [newick.add_nodes(fm) for fm in self.files.values()]

        # Determine the counts for the common taxa.
        d_rank_common = self.get_n_common()

        # Create an ete3 tree and annotate it
        t = ete3.Tree(str(newick), format=1, quoted_node_names=True)
        ts = ete3.TreeStyle()
        ts.show_leaf_name = False
        ts.show_scale = False
        ts.rotation = rotation_deg

        def my_layout(node):
            COLOURS = ['#ADEF29', '#F0E442', '#009E73', '#56B4E9', '#E69F00', '#911eb4', '#46f0f0', '#f032e6',
                       '#bcf60c', '#fabebe', '#008080', '#e6beff', '#9a6324', '#fffac8', '#800000', '#aaffc3',
                       '#808000', '#ffd8b1', '#000075', '#808080', '#ffffff', '#000000']

            F = ete3.TextFace(node.name, fsize=8, tight_text=True)
            F.rotation = -rotation_deg
            F.margin_right = 2
            F.margin_left = 3
            ete3.add_face_to_node(F, node, column=0, position='branch-right')

            tf_invis = FMeasureTree.get_text_face('-', None, 0.0)

            if node.name in ['d__Bacteria', 'd__Archaea']:

                # Attach a legend to the plot.
                if legend:
                    tf_legend_common_in = FMeasureTree.get_text_face('No. Common In', '#F2F2F2', 0.5, fsize=7,
                                                                     bold=True)
                    ete3.faces.add_face_to_node(tf_legend_common_in, node, column=1, position="branch-top")

                    tf_legend_common_out = FMeasureTree.get_text_face('No. Common Out', '#F2F2F2', 0.5, fsize=7,
                                                                      bold=True)
                    ete3.faces.add_face_to_node(tf_legend_common_out, node, column=1, position="branch-top")

                    tf_legend_common_blank = FMeasureTree.get_text_face('X', None, 0.0, fsize=7)
                    ete3.faces.add_face_to_node(tf_legend_common_blank, node, column=1, position="branch-top")

                    for idx, model_id in enumerate(sorted(self.files.keys())):
                        tf_legend_in = FMeasureTree.get_text_face(f'No. Rogue In ({model_id})', COLOURS[idx], 1.0,
                                                                  fsize=7)
                        ete3.faces.add_face_to_node(tf_legend_in, node, column=2 + idx, position="branch-top")

                        tf_legend_out = FMeasureTree.get_text_face(f'No. Rogue Out ({model_id})', COLOURS[idx], 1.0,
                                                                   fsize=7)
                        ete3.faces.add_face_to_node(tf_legend_out, node, column=2 + idx, position="branch-top")

                        tf_legend_exp = FMeasureTree.get_text_face(f'No. Expected ({model_id})', "#F2F2F2", 0.2,
                                                                   fsize=7)
                        ete3.faces.add_face_to_node(tf_legend_exp, node, column=2 + idx, position="branch-top")

            # Check if this node was not monophyletic in any of the models
            poly_ranks = self.get_poly_ranks()
            if node.name in poly_ranks:

                # Create a spacer between the rank name and the values.
                ete3.faces.add_face_to_node(tf_invis, node, column=1, position="branch-right")

                n_common_in = d_rank_common[node.name]['in']
                n_common_out = d_rank_common[node.name]['out']

                # Add a spacer to act as a top margin.
                ete3.faces.add_face_to_node(tf_invis, node, column=2, position="branch-right")

                # Add the number of common taxa in.
                if n_common_in == 0:
                    tf_common_in = FMeasureTree.get_text_face(n_common_in, '#f2f2f2', 0.2, bold=True)
                else:
                    tf_common_in = FMeasureTree.get_text_face(n_common_in, '#f2f2f2', 0.5, bold=True)
                ete3.faces.add_face_to_node(tf_common_in, node, column=2, position="branch-right")

                # Add the number of common taxa out.
                if n_common_out == 0:
                    tf_common_out = FMeasureTree.get_text_face(n_common_out, '#f2f2f2', 0.2, bold=True)
                else:
                    tf_common_out = FMeasureTree.get_text_face(n_common_out, '#f2f2f2', 0.5, bold=True)
                ete3.faces.add_face_to_node(tf_common_out, node, column=2, position="branch-right")

                tf_common_blank = FMeasureTree.get_text_face('X', None, 0.0)
                ete3.faces.add_face_to_node(tf_common_blank, node, column=2, position="branch-right")

                # Add model specific information for this current rank.
                for idx, model_id in enumerate(sorted(self.files.keys())):
                    cur_f_table = self.files[model_id].content
                    n_rogue_in = len(cur_f_table[node.name]["rogue_in"])
                    n_rogue_out = len(cur_f_table[node.name]["rogue_out"])

                    # Add a top margin spacer.
                    ete3.faces.add_face_to_node(tf_invis, node, column=3 + idx, position="branch-right")

                    # Number of rogue taxa in.
                    if n_rogue_in == 0:
                        tf_generic_in = FMeasureTree.get_text_face(n_rogue_in, '#f2f2f2', 0.2)
                    else:
                        tf_generic_in = FMeasureTree.get_text_face(n_rogue_in, COLOURS[idx], 1.0)
                    ete3.faces.add_face_to_node(tf_generic_in, node, column=3 + idx, position="branch-right")

                    # Number of rogue taxa out.
                    if n_rogue_out == 0:
                        tf_generic_out = FMeasureTree.get_text_face(n_rogue_out, '#f2f2f2', 0.2)
                    else:
                        tf_generic_out = FMeasureTree.get_text_face(n_rogue_out, COLOURS[idx], 1.0)
                    ete3.faces.add_face_to_node(tf_generic_out, node, column=3 + idx, position="branch-right")

                    # Expected number of taxa in this model.
                    tf_expected_cnt = FMeasureTree.get_text_face(cur_f_table[node.name]["n_expected"], '#f2f2f2', 0.2)
                    ete3.faces.add_face_to_node(tf_expected_cnt, node, column=3 + idx, position="branch-right")

        ts.layout_fn = my_layout

        if out_path is None:
            print('Run as: xvfb-run python concat_f_measure.py')
            t.show(tree_style=ts)
        else:
            t.render(out_path, tree_style=ts)


class TaxonomyFile(object):

    def __init__(self, path):
        self.path = path
        self.contents = self.read()

    def read(self):
        out = dict()
        with open(self.path, 'r') as f:
            for line in f.readlines():
                cols = line.strip().split('\t')
                accession = cols[0]
                ranks = cols[1].split(';')
                out[accession] = ranks
        return out

    def get_content(self):
        return self.contents

    def get_taxon_namespace(self):
        out = set()

        for v in self.contents.values():
            [out.add(x) for x in v]

        return out

    def get_ranks_above(self, search_str):
        for v in self.contents.values():
            last_ranks = list()
            for rank in v:
                last_ranks.append(rank)
                if rank == search_str:
                    return last_ranks


class Node(object):

    def __init__(self, name, parent):
        self.name = name
        self.children = list()
        self.parent = parent
        self.count = None
        self.metadata = list()

    def get_name(self):
        return self.name

    def add_child(self, node):
        self.children.append(node)

    def get_child(self, name):
        for child in self.children:
            if child.get_name() == name:
                return child
        return None

    def add_ranks(self, ranks):
        last_node = self
        for rank in ranks:
            search_node = last_node.get_child(rank)
            if not search_node:
                new_node = Node(rank, parent=last_node)
                last_node.add_child(new_node)
                last_node = new_node
            else:
                last_node = search_node
        return last_node

    def add_expected_count(self, count):
        self.count = count

    def __str__(self):

        parent_str = 'null' if self.parent is None else self.parent.get_name()

        expected_str = '' if self.count is None else ',\n"n_expected": %s' % self.count
        metadata_str = '' if len(self.metadata) == 0 else ',\n"metadata": [\n%s]' % '\n'.join(self.metadata)

        children_str = str()
        if len(self.children) > 0:
            children_str = ''',
            "children": [
            %s
            ]''' % ', '.join([str(x) for x in self.children])

        out = '''{
        "name": "%s",
        "parent": "%s" %s %s %s 
    }
        ''' % (self.name, parent_str, children_str, expected_str, metadata_str)

        return out

    def add_metadata(self, item):
        self.metadata.append(item)


class FMeasureTable(object):
    cols = ('Taxon', 'No. Expected in Tree', 'F-measure', 'Precision', 'Recall',
            'No. Genomes from Taxon', 'No. Genome In Lineage', 'Rogue out',
            'Rogue in')

    def __init__(self, path):
        self.path = path
        self.content = self.read()

    def read(self):
        out = dict()
        with open(self.path) as fh:
            read_cols = tuple([x for x in fh.readline().strip().split('\t')])
            if self.cols != read_cols:
                raise Exception('PhyloRank output file has different headers.')
            col_ids = {x: i for i, x in enumerate(self.cols)}

            for line in fh.readlines():
                vals = [x.strip() for x in line.split('\t')]

                taxon = vals[col_ids['Taxon']]
                n_expected = int(vals[col_ids['No. Expected in Tree']])
                f_measure = float(vals[col_ids['F-measure']])
                precision = float(vals[col_ids['Precision']])
                recall = float(float(vals[col_ids['Recall']]))
                n_from_taxon = int(vals[col_ids['No. Genomes from Taxon']])
                n_from_lineage = int(vals[col_ids['No. Genome In Lineage']])
                rogue_in = vals[col_ids['Rogue in']]
                rogue_out = vals[col_ids['Rogue out']]

                hit = dict()
                hit['n_expected'] = n_expected
                hit['f_measure'] = f_measure
                hit['precision'] = precision
                hit['recall'] = recall
                hit['n_from_taxon'] = n_from_taxon
                hit['n_from_lineage'] = n_from_lineage
                hit['rogue_in'] = list() if len(rogue_in) == 0 else rogue_in.split(',')
                hit['rogue_out'] = list() if len(rogue_out) == 0 else rogue_out.split(',')
                out[taxon] = hit
        return out

    def get_content(self) -> dict:
        return self.content


class NewickTree(object):

    def __init__(self, tf: TaxonomyFile):
        self.tf = tf
        self.taxon_namespace = dendropy.TaxonNamespace(self.tf.get_taxon_namespace())
        self.tree = dendropy.Tree()
        self.root = self.tree.seed_node

    def add_nodes(self, fm: FMeasureTable):
        for taxon, f_dict in fm.get_content().items():

            # Only constructing nodes which are polyphyletic.
            if f_dict['f_measure'] >= 1.0:
                continue

            last_node = self.root
            for cur_rank in self.tf.get_ranks_above(taxon):
                cur_node = self.tree.find_node_with_taxon_label(cur_rank)
                if not cur_node:
                    if cur_rank in ['d__Archaea', 'd__Bacteria']:
                        self.root.taxon = self.taxon_namespace.get_taxon(cur_rank)
                        cur_node = self.root
                    else:
                        new_node = dendropy.Node()
                        last_node.add_child(new_node)
                        new_node.taxon = self.taxon_namespace.get_taxon(cur_rank)
                        cur_node = new_node
                last_node = cur_node

    def __str__(self):
        return self.tree.as_string('newick')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import argparse
import logging
import sys
import traceback

from metatree import __description__, __version__
from metatree.common import check_on_path
from metatree.exception import MetaTreeException, MetaTreeExit
from metatree.io import Batchfile
from metatree.io.taxonomy_file import TaxonomyFile
from metatree.logger import logger_setup
from metatree.pipeline import run_pipeline


def print_help():
    lines = [f'metatree v{__version__}',
             'usage: [batchfile] [out_dir] [taxonomy_file] [outgroup] [cpus]']
    print('\n'.join(lines))


def main(args=None):
    parser = argparse.ArgumentParser(description=__description__)
    parser.add_argument('batchfile', type=str,
                        help='First tree must be the reference tree, format: id<tab>path_to_tree')
    parser.add_argument('out_dir', type=str, help='path to the output directory')
    parser.add_argument('taxonomy_file', type=str, help='path to taxonomy file, format: gid<tab>taxonomy')
    parser.add_argument('outgroup', type=str, help='outgroup for rooting')
    parser.add_argument('cpus', type=int, help='number of CPUs to use')

    # Verify that a subparser was selected
    if len(sys.argv) == 1:
        print_help()
        sys.exit(0)
    elif sys.argv[1] in {'-v', '--v', '-version', '--version'}:
        print(f'metatree v{__version__}')
        sys.exit(0)
    else:
        print(f'metatree v{__version__}')
        args = parser.parse_args(args)

        # Setup the logger.
        logger_setup(args.out_dir if hasattr(args, 'out_dir') else None,
                     "metatree.log", "metatree", __version__, False,
                     hasattr(args, 'debug') and args.debug)
        logger = logging.getLogger('timestamp')

        try:
            # Validate the input arguments.
            batchfile = Batchfile(args.batchfile)
            tax_file = TaxonomyFile(args.taxonomy_file)
            if not args.outgroup or args.outgroup[0:3] not in {'d__', 'p__', 'c__', 'o__', 'f__', 'g__', 's__'}:
                raise MetaTreeExit(f'Invalid outgroup: {args.outgroup}')
            cpus = max(1, args.cpus)

            # Assert that the required programs are on the system path.
            for prog in ('genometreetk', 'phylorank'):
                check_on_path(prog)

            # Run the pipeline.
            run_pipeline(batchfile, args.out_dir, tax_file, args.outgroup, cpus)

        except SystemExit:
            sys.stdout.write('\n')
            sys.stdout.flush()
            logger.error('Controlled exit resulting from early termination.')
            sys.exit(1)
        except KeyboardInterrupt:
            sys.stdout.write('\n')
            sys.stdout.flush()
            logger.error('Controlled exit resulting from interrupt signal.')
            sys.exit(1)
        except MetaTreeExit as e:
            sys.stdout.write('\n')
            sys.stdout.flush()
            if len(str(e)) > 0:
                logger.error('{}'.format(e))
            logger.error('Controlled exit resulting from an unrecoverable error or warning.')
            sys.exit(1)
        except MetaTreeException as e:
            sys.stdout.write('\n')
            sys.stdout.flush()
            msg = 'Controlled exit resulting from an unrecoverable error or warning.\n\n'
            msg += '=' * 80 + '\n'
            msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
            msg += '  MESSAGE: {}\n'.format(e)
            msg += '_' * 80 + '\n\n'
            msg += traceback.format_exc()
            msg += '=' * 80
            logger.error(msg)
            sys.exit(1)
        except Exception as e:
            sys.stdout.write('\n')
            sys.stdout.flush()
            msg = 'Uncontrolled exit resulting from an unexpected error.\n\n'
            msg += '=' * 80 + '\n'
            msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
            msg += '  MESSAGE: {}\n'.format(e)
            msg += '_' * 80 + '\n\n'
            msg += traceback.format_exc()
            msg += '=' * 80
            logger.error(msg)
            sys.exit(1)

    # Done - no errors.
    logger.info('Done.')
    sys.exit(0)


if __name__ == "__main__":
    main()



import logging
import os
from collections import defaultdict

import pandas as pd

from metatree.io import Batchfile


class FMeasureTable(object):

    def __init__(self, path):
        self.path = path
        self.content = self.read()

    def read(self):
        out = dict()

        df = pd.read_csv(self.path, delimiter='\t')
        df_filtered = df[df['F-measure'] < 1.0]

        for idx, row in df_filtered.iterrows():
            hit = dict()
            hit['n_expected'] = row['No. Expected in Tree']
            hit['n_in'] = 0 if str(row['Rogue in']) == 'nan' else len(row['Rogue in'].split(','))
            hit['g_in'] = list() if str(row['Rogue in']) == 'nan' else row['Rogue in'].split(',')
            hit['n_out'] = 0 if str(row['Rogue out']) == 'nan' else len(row['Rogue out'].split(','))
            hit['g_out'] = list() if str(row['Rogue out']) == 'nan' else row['Rogue out'].split(',')
            out[row['Taxon']] = hit

        return out

    def get_content(self) -> dict:
        return self.content


def parse_tax_file(path):
    out = dict()
    with open(path, 'r') as f:
        for line in f.readlines():
            gid, tax = line.strip().split('\t')
            out[gid] = tax
    return out


def tax_to_dict(tax):
    out = defaultdict(list)
    for rank in tax.split(';'):
        out[rank[0]].append(rank)
    return out


def compare_tax(truth, model, comp):
    n_disagree = 0
    n_poly = 0
    n_agree = 0

    for gid, m_tax in comp.items():
        t_tax = truth[gid]

        t_tax, m_tax = t_tax.replace('; ', ';'), m_tax.replace('; ', ';')
        dict_t_tax, dict_m_tax = tax_to_dict(t_tax), tax_to_dict(m_tax)

        # Case 1: Agreement
        if m_tax == t_tax:
            n_agree += 1
            continue

        for rank in ['d', 'p', 'c', 'o', 'f', 'g', 's']:
            m_rank, t_rank = dict_m_tax[rank], dict_t_tax[rank]

            if m_rank == t_rank:
                continue

            # Case 2: Disagreement
            elif len(m_rank) == len(t_rank):
                n_disagree += 1

                print(t_tax)
                print(m_tax)
                print('--')

                break

            # Case 3: Polyphyletic / novel rank
            else:
                n_poly += 1
                break

    print(f'{model}\t{n_agree}\t{n_disagree}\t{n_poly}')
    return model, n_agree, n_disagree, n_poly


class MismatchTable(object):

    def __init__(self, path):
        self.path = path
        self.logger = logging.getLogger('timestamp')

    def run_and_save(self, batchfile: Batchfile, dir_decorated, path_tax):
        truth_tax = parse_tax_file(path_tax.path)

        for tree_id, tree_path in batchfile.data.items():
            cur_tax = parse_tax_file(os.path.join(dir_decorated, f'{tree_id}_rooted_decorated.tree-taxonomy'))
            compare_tax(truth_tax, tree_id, cur_tax)

        return



import logging
import os

from metatree.common import make_sure_path_exists
from metatree.exception import MetaTreeExit


class RfResults(object):

    def __init__(self, path):
        self.logger = logging.getLogger('timestamp')
        self.path = path
        self.data = self.read()
        make_sure_path_exists(os.path.dirname(path))

    def read(self):
        out = dict()
        if os.path.isfile(self.path):
            with open(self.path) as fh:
                for line in fh.readlines():
                    tid_a, tid_b, rf, norm_rf = line.strip().split('\t')
                    out[(tid_a, tid_b)] = (float(rf), float(norm_rf))
        return out

    def is_done(self, tid_a, tid_b):
        return (tid_a, tid_b) in self.data or (tid_b, tid_a) in self.data

    def add(self, tid_a, tid_b, rf, norm_rf):
        self.data[(tid_a, tid_b)] = (rf, norm_rf)

    def write(self):
        done = dict()
        with open(self.path, 'w') as fh:
            for (tid_a, tid_b), (rf, norm_rf) in self.data.items():
                if (tid_a, tid_b) in done and self.data[(tid_a, tid_b)] != done[(tid_a, tid_b)]:
                    raise MetaTreeExit('Inconsistent results, report this issue.')
                if (tid_b, tid_a) in done and self.data[(tid_b, tid_a)] != done[(tid_b, tid_a)]:
                    raise MetaTreeExit('Inconsistent results, report this issue.')
                if (tid_a, tid_b) in done or (tid_b, tid_a) in done:
                    continue
                else:
                    fh.write(f'{tid_a}\t{tid_b}\t{rf}\t{norm_rf}\n')
                    done[(tid_a, tid_b)] = (rf, norm_rf)
        self.logger.info(f'Pairwise Robinson-Foulds distances written to: {self.path}')



import logging
import os
import subprocess
from multiprocessing import Pool

from genometreetk import __version__ as genometreetk_v
from tqdm import tqdm

from metatree.common import make_sure_path_exists
from metatree.exception import MetaTreeExit
from metatree.io import Batchfile
from metatree.io.taxonomy_file import TaxonomyFile


class TreeRoot(object):

    def __init__(self, dir_root):
        self.dir_root = dir_root
        self.logger = logging.getLogger('timestamp')
        make_sure_path_exists(dir_root)

    @staticmethod
    def worker(task):
        tree_id, tree_in, tree_out, outgroup, tax_file = task
        args = ['genometreetk', 'outgroup', tree_in, tax_file, outgroup, tree_out]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        proc.communicate()
        if not proc.returncode == 0:
            raise MetaTreeExit(f'Non-zero return code: {" ".join(args)}')

    def run(self, batchfile: Batchfile, dir_root: str, outgroup: str, tax_file: TaxonomyFile, cpus: int):
        queue = list()
        for tree_id, tree_in in batchfile.data.items():
            tree_out = os.path.join(dir_root, f'{tree_id}_rooted.tree')
            if not os.path.isfile(tree_out):
                queue.append((tree_id, tree_in, tree_out, outgroup, tax_file.path))

        self.logger.info(f'Rooting trees using GenomeTreeTk v{genometreetk_v}')
        with Pool(processes=cpus) as pool:
            for _ in tqdm(pool.imap_unordered(TreeRoot.worker, queue), total=len(queue)):
                pass



import logging
import os
import tempfile
from collections import defaultdict
from multiprocessing import Pool
from warnings import simplefilter

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from Bio import Phylo
from Bio.Phylo.TreeConstruction import DistanceMatrix, DistanceTreeConstructor
from scipy.cluster.hierarchy import ClusterWarning
from tqdm import tqdm

from metatree.external.tree_compare import TreeCompare
from metatree.io import Batchfile, RfResults

simplefilter("ignore", ClusterWarning)


class TreeDist(object):

    def __init__(self):
        self.logger = logging.getLogger('timestamp')

    @staticmethod
    def worker(task):
        tid_a, path_a, tid_b, path_b, set_common = task

        if set_common:
            with tempfile.TemporaryDirectory() as tmp_dir:
                taxa_list = os.path.join(tmp_dir, 'taxa_list.tsv')
                with open(taxa_list, 'w') as fh:
                    for taxa in set_common:
                        fh.write(f'{taxa}\n')

                # Calculate the RF distance.
                tc = TreeCompare()
                rf, norm_rf = tc.robinson_foulds(path_a, path_b, taxa_list)

        else:
            # Calculate the RF distance.
            tc = TreeCompare()
            rf, norm_rf = tc.robinson_foulds(path_a, path_b, None)

        return tid_a, tid_b, rf, norm_rf

    def run(self, rf_results: RfResults, batchfile: Batchfile, dir_root, dir_dec, cpus: int, common_taxa: bool):

        # Determine if a common subset of taxa should be used.
        if common_taxa:
            set_common = batchfile.common_taxa()
            self.logger.info(f'Robinson-Foulds metrics will only consider those {len(set_common):,} '
                             f'taxa which are common between ALL trees.')
        else:
            set_common = None

        # Determine which trees still need to be processed.
        queue = list()
        tree_ids = list(batchfile.data.keys())
        for i in range(len(tree_ids)):
            tid_a = tree_ids[i]
            path_a = batchfile.data[tid_a]
            for j in range(i):
                tid_b = tree_ids[j]
                path_b = batchfile.data[tid_b]
                if not rf_results.is_done(tid_a, tid_b):
                    queue.append((tid_a, path_a, tid_b, path_b, set_common))

        self.logger.info(f'Calculating Robinson-Foulds distances.')
        with Pool(processes=cpus) as pool:
            for tid_a, tid_b, rf, norm_rf in tqdm(pool.imap_unordered(TreeDist.worker, queue), total=len(queue)):
                rf_results.add(tid_a, tid_b, rf, norm_rf)

        rf_results.write()

    def summarise_dist(self, rf_results: RfResults, dir_out):

        for use_norm in (True, False):
            if use_norm:
                path_out = os.path.join(dir_out, 'rf_normed.tree')
                path_hm = os.path.join(dir_out, 'rf_normed_heatmap.svg')
                plt_title = 'Normalised Robinson-Foulds Distance'
            else:
                path_out = os.path.join(dir_out, 'rf_un_normed.tree')
                path_hm = os.path.join(dir_out, 'rf_un_normed_heatmap.svg')
                plt_title = '(un)Normalised Robinson-Foulds Distance'

            metrics = defaultdict(dict)
            names = set()
            for (tid_a, tid_b), (rf, norm_rf) in rf_results.data.items():
                if use_norm:
                    metrics[tid_a][tid_b] = norm_rf
                    metrics[tid_b][tid_a] = norm_rf
                else:
                    metrics[tid_a][tid_b] = rf
                    metrics[tid_b][tid_a] = rf
                names.add(tid_a)
                names.add(tid_b)

            labels = sorted(list(names))
            mat_vals = list()
            mat = np.zeros((len(labels), len(labels)))
            for i in range(len(labels)):
                cur_row = list()
                tid_a = labels[i]
                for j in range(i + 1):
                    tid_b = labels[j]
                    if tid_a == tid_b:
                        cur_row.append(0.0)
                    else:
                        cur_row.append(metrics[tid_a][tid_b])
                        mat[i, j] = metrics[tid_a][tid_b]
                mat_vals.append(cur_row)
            mat = mat + mat.T

            # Newick
            dm = DistanceMatrix(names=labels, matrix=mat_vals)
            constructor = DistanceTreeConstructor()
            tree = constructor.nj(dm)

            Phylo.write(tree, path_out, 'newick')

            # Heatmap
            cmap = sns.cubehelix_palette(100, reverse=True)

            sns.set(font_scale=1)
            fig_size = (15, 15)

            rf_df = pd.DataFrame(mat, columns=labels, index=labels)
            sns.clustermap(rf_df, annot=True, fmt='.3f', cmap=cmap, figsize=fig_size).fig.suptitle(plt_title)
            plt.savefig(path_hm)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2017'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__version__ = '0.0.1'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'
__status__ = 'Development'

import logging
import sys

import dendropy
from biolib.newick import parse_label, create_label
from dendropy.calculate import treecompare


class TreeCompare(object):
    """Compare pairs of trees."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger('timestamp')

    def _check_fractional_bootstraps(self, tree):
        """Check if bootstrap values are between [0, 1] and change to [0, 100]."""

        fractional_bootstrap = True
        for n in tree.preorder_node_iter():
            support, label, aux_info = parse_label(n.label)
            if support is not None and support > 1.0:
                fractional_bootstrap = False
                break

        if fractional_bootstrap:
            for n in tree.preorder_node_iter():
                support, label, aux_info = parse_label(n.label)
                if support is not None:
                    n.label = create_label(int(support * 100 + 0.5), label, aux_info)

    def _prune(self, tree1, tree2):
        """Prune trees to common set of taxa."""

        # prune both trees to the set of common taxa
        taxa1 = set([t.taxon for t in tree1.leaf_node_iter()])
        taxa2 = set([t.taxon for t in tree2.leaf_node_iter()])

        if len(taxa1) != len(taxa2) or taxa1 != taxa2:
            taxa_in_common = taxa1.intersection(taxa2)
            # self.logger.info('Tree 1 contains %d taxa.' % len(taxa1))
            # self.logger.info('Tree 2 contains %d taxa.' % len(taxa2))
            # self.logger.info('Pruning trees to the %d taxa in common.' % len(taxa_in_common))

            if not taxa_in_common:
                self.logger.error('No taxa in common.')
                sys.exit(-1)

            tree1.retain_taxa(taxa_in_common)
            tree2.retain_taxa(taxa_in_common)

    def _read_trees(self, tree1_file, tree2_file, taxa_list, prune=True):
        """Read trees from file."""

        tns = dendropy.TaxonNamespace()
        tree1 = dendropy.Tree.get_from_path(tree1_file,
                                            schema='newick',
                                            rooting='force-unrooted',
                                            preserve_underscores=True,
                                            taxon_namespace=tns)

        tree2 = dendropy.Tree.get_from_path(tree2_file,
                                            schema='newick',
                                            rooting='force-unrooted',
                                            preserve_underscores=True,
                                            taxon_namespace=tns)

        # check if bootstrap values are fractional
        self._check_fractional_bootstraps(tree1)
        self._check_fractional_bootstraps(tree2)

        # prune trees to specified taxa
        if taxa_list:
            taxa_to_keep = set()
            for line in open(taxa_list):
                taxa_to_keep.add(line.strip().split('\t')[0])
            tree1.retain_taxa_with_labels(taxa_to_keep)
            tree2.retain_taxa_with_labels(taxa_to_keep)

        # verify trees were defined over the same set of taxa
        if prune:
            self._prune(tree1, tree2)

        tree1.encode_bipartitions()
        tree2.encode_bipartitions()

        assert tree1.taxon_namespace is tree2.taxon_namespace

        return tree1, tree2

    def euclidean(self, tree1, tree2, taxa_list):
        """Calculate Euclidean distance between two trees."""

        tree1, tree2 = self._read_trees(tree1, tree2, taxa_list)

        return treecompare.euclidean_distance(tree1, tree2)

    def robinson_foulds(self, tree1, tree2, taxa_list):
        """Calculate Robinson-Foulds (i.e., symmetric_difference) distance between two trees."""

        tree1, tree2 = self._read_trees(tree1, tree2, taxa_list)

        rf = treecompare.symmetric_difference(tree1, tree2)

        num_taxa = len([t for t in tree1.leaf_node_iter()])
        normalized_rf = float(rf) / (2 * (num_taxa - 3))

        return rf, normalized_rf

    def weighted_robinson_foulds(self, tree1, tree2, taxa_list):
        """Calculate weighted Robinson-Foulds distance between two trees."""

        tree1, tree2 = self._read_trees(tree1, tree2, taxa_list)

        return treecompare.weighted_robinson_foulds_distance(tree1, tree2)

    def report_all(self, tree1_file, tree2_file):
        """Report results for all tree comparison statistics."""

        tree1, tree2 = self._read_trees(tree1_file, tree2_file)

        print('Euclidean: %f' % self.euclidean(tree1, tree2))
        print('Robinson-Foulds: %f' % self.robinson_foulds(tree1, tree2))
        print('Weighted Robinson-Foulds: %f' % self.weighted_robinson_foulds(tree1, tree2))

    def report_missing_splits(self, ref_tree, compare_tree, min_support, taxa_list):
        """Report supported bipartitions in reference tree not in comparison tree."""

        ref_tree, compare_tree = self._read_trees(ref_tree, compare_tree, taxa_list)

        incongruent = 0
        print('Missing splits with support >= %f:' % min_support)
        for n in ref_tree.preorder_node_iter(lambda n: not n.is_leaf()):
            support, label, aux_info = parse_label(n.label)

            if support >= min_support:
                if n.bipartition not in compare_tree.bipartition_encoding:
                    incongruent += 1
                    if label:
                        print(label, n.edge.length)
                    else:
                        print(','.join([t.taxon.label for t in n.leaf_iter()]))

        print('Missing splits: %d' % incongruent)

    def _avg_descendant_rate(self, tree):
        """Calculate average rate of divergence for each nodes in a tree.

        The average rate is the arithmetic mean of the
        branch length to all descendant taxa.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        The following attributes are added to each node:
          mean_dist: mean distance to tips
          num_taxa: number of terminal taxa
        """

        # calculate the mean branch length to extant taxa
        for node in tree.postorder_node_iter():
            avg_div = 0
            if node.is_leaf():
                node.mean_dist = 0.0
                node.num_taxa = 1
            else:
                node.num_taxa = sum([1 for _ in node.leaf_iter()])
                for c in node.child_node_iter():
                    num_tips = c.num_taxa
                    avg_div += (float(c.num_taxa) / node.num_taxa) * (c.mean_dist + c.edge_length)

            node.mean_dist = avg_div

    def _rel_dist(self, tree):
        """Calculate relative distance to each internal node.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        The following attributes are added to each node:
          mean_dist: mean distance to tips
          num_taxa: number of terminal taxa
          rel_dists: relative distance of node between root and extant organisms
        """

        self._avg_descendant_rate(tree)

        for node in tree.preorder_node_iter():
            if node == tree.seed_node:
                node.rel_dist = 0.0
            elif node.is_leaf():
                node.rel_dist = 1.0
            else:
                a = node.edge_length
                b = node.mean_dist
                x = node.parent_node.rel_dist

                if (a + b) != 0:
                    rel_dist = x + (a / (a + b)) * (1.0 - x)
                else:
                    # internal node has zero length to parent,
                    # so should have the same relative distance
                    # as the parent node
                    rel_dist = x

                node.rel_dist = rel_dist

    def _supported(self, ref_tree, compare_tree, min_support, max_depth):
        """Determine supported bipartitions in reference tree not in comparison tree."""

        congruent = 0
        congruent_w = 0
        incongruent = 0
        incongruent_w = 0
        nontrivial_splits = 0
        nontrivial_splits_w = 0
        congruent_splits = {}
        incongruent_splits = {}
        for n in ref_tree.preorder_node_iter(lambda n: not n.is_leaf()):
            if not n.parent_node:
                continue

            nontrivial_splits += 1
            nontrivial_splits_w += n.edge.length

            support, label, aux_info = parse_label(n.label)
            if support is None or (support >= min_support and n.rel_dist <= max_depth):
                split_lca = n.child_nodes()[0].leaf_nodes()[0].taxon.label
                split_lca += '|'
                split_lca += n.child_nodes()[1].leaf_nodes()[0].taxon.label
                if n.bipartition not in compare_tree.bipartition_encoding:
                    incongruent += 1
                    incongruent_w += n.edge.length
                    incongruent_splits[split_lca] = (n.edge.length, support)
                else:
                    congruent += 1
                    congruent_w += n.edge.length
                    congruent_splits[split_lca] = (n.edge.length, support)

        return congruent, congruent_w, incongruent, incongruent_w, nontrivial_splits, nontrivial_splits_w, congruent_splits, incongruent_splits

    def _shared_support(self, tree1, tree2, min_support, max_depth):
        """Determine supported bipartitions common to a pair of trees."""

        assert tree1.taxon_namespace is tree2.taxon_namespace

        common_supported_splits = 0
        common_supported_splits_w = 0
        null_support = False
        for n in tree1.preorder_node_iter(lambda n: not n.is_leaf()):
            if not n.parent_node:
                continue

            support, label, aux_info = parse_label(n.label)
            if support is None:
                null_support = True

            if support is None or (support >= min_support and n.rel_dist <= max_depth):
                if n.bipartition in tree2.bipartition_encoding:
                    edge2 = tree2.bipartition_edge_map[n.bipartition]
                    support2, label2, aux_info2 = parse_label(edge2.head_node.label)

                    if support2 is None or (support2 >= min_support and edge2.head_node.rel_dist <= max_depth):
                        common_supported_splits += 1
                        common_supported_splits_w += n.edge.length + edge2.length

        if null_support:
            self.logger.warning('Some internal nodes lack support values and were treated as supported.')

        return common_supported_splits, common_supported_splits_w

    def supported_splits(self, tree1_file, tree2_file, split_file, min_support, max_depth, taxa_list):
        """Supported bipartitions of common taxa shared between two trees."""

        tree1, tree2 = self._read_trees(tree1_file, tree2_file, taxa_list)

        self._rel_dist(tree1)
        self._rel_dist(tree2)

        r = self._supported(tree1, tree2, min_support, max_depth)
        (congruent12, congruent12_w,
         incongruent12, incongruent12_w,
         nontrivial_splits1, nontrivial_splits1_w,
         congruent_splits12, incongruent_splits12) = r

        r = self._supported(tree2, tree1, min_support, max_depth)
        (congruent21, congruent21_w,
         incongruent21, incongruent21_w,
         nontrivial_splits2, nontrivial_splits2_w,
         congruent_splits21, incongruent_splits21) = r

        common_supported_splits, common_supported_splits_w = self._shared_support(tree1, tree2, min_support, max_depth)

        print('')
        print('T1 -> T2')
        print('Non-trivial splits in T1: %d' % nontrivial_splits1)
        print('Weight of non-trivial splits in T1: %g' % nontrivial_splits1_w)
        print('Well-supported splits in T1: %d' % (congruent12 + incongruent12))
        print('Weight of well-supported splits in T1: %g' % (congruent12_w + incongruent12_w))
        print('Well-supported splits in T1 present in T2: %d' % congruent12)
        print('Weight of well-support splits in T1 present in T2: %g' % congruent12_w)

        print('')
        print('T2 -> T1')
        print('Non-trivial splits in T2: %d' % nontrivial_splits2)
        print('Weight of non-trivial splits in T2: %g' % nontrivial_splits2_w)
        print('Well-supported splits in T2: %d' % (congruent21 + incongruent21))
        print('Weight of well-supported splits in T2: %g' % (congruent21_w + incongruent21_w))
        print('Well-supported splits in T2 present in T1: %d' % congruent21)
        print('Weight of well-support splits in T2 present in T1: %g' % congruent21_w)

        print('')
        print('T1 <-> T2')
        print('Non-trivial splits in T1 and T2: %d' % (nontrivial_splits1 + nontrivial_splits2))
        print('Weight of non-trivial splits in T1 and T2: %g' % (nontrivial_splits1_w + nontrivial_splits2_w))
        print('Well-supported splits in T1 and T2: %d' % (congruent12 + incongruent12 + congruent21 + incongruent21))
        supported_splits_t12 = congruent12_w + incongruent12_w + congruent21_w + incongruent21_w
        print('Weight of well-supported splits in T1 and T2: %g' % supported_splits_t12)
        print('Well-supported splits in common between T1 and T2: %d' % common_supported_splits)
        print('Weight of well-support splits in common between T1 and T2: %g' % common_supported_splits_w)
        print('Fraction of common well-supported splits: %g' % (common_supported_splits_w / supported_splits_t12))

        if split_file:
            fout = open(split_file, 'w')
            fout.write('Split\tStatus\tTree Comparison\tBranch Length\tSupport\n')
            for split_lca, (length, support) in congruent_splits12.items():
                if support is not None:
                    fout.write('%s\t%s\t%s\t%f\t%.2f\n' % (split_lca, 'Congruent', 'T1 -> T2', length, support))
            for split_lca, (length, support) in congruent_splits21.items():
                if support is not None:
                    fout.write('%s\t%s\t%s\t%f\t%.2f\n' % (split_lca, 'Congruent', 'T2 -> T1', length, support))
            for split_lca, (length, support) in incongruent_splits12.items():
                if support is not None:
                    fout.write('%s\t%s\t%s\t%f\t%.2f\n' % (split_lca, 'Incongruent', 'T1 -> T2', length, support))
            for split_lca, (length, support) in incongruent_splits21.items():
                if support is not None:
                    fout.write('%s\t%s\t%s\t%f\t%.2f\n' % (split_lca, 'Incongruent', 'T2 -> T1', length, support))
            fout.close()



import os
import re

from setuptools import setup


def read_meta():
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'metatree/__init__.py')
    with open(path) as fh:
        hits = re.findall(r'__(\w+)__ ?= ?["\'](.+)["\']\n', fh.read())
    return {k: v for k, v in hits}


def readme():
    with open('README.md') as f:
        return f.read()


meta = read_meta()

setup(name=meta['title'],
      version=meta['version'],
      description=meta['description'],
      author=meta['author'],
      author_email=meta['author_email'],
      url=meta['url'],
      license=meta['license'],
      long_description=readme(),
      long_description_content_type='text/markdown',
      project_urls={
          "Bug Tracker": "https://github.com/aaronmussig/PhyloDM/issues",
          "Documentation": "https://github.com/aaronmussig/PhyloDM",
          "Source Code": "https://github.com/aaronmussig/PhyloDM",
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='phylogenetic distance matrix symmetric',
      packages=['metatree', 'metatree.io', 'metatree.external'],
      entry_points={
          'console_scripts': [
              'metatree = metatree.__main__:main'
          ]
      },
      install_requires=['phylorank>0.1.0', 'genometreetk>0.1.2', 'dendropy>=4.1.0', 'tqdm>=4.31.0', 'biolib>=0.1.0',
                        'biopython', 'seaborn', 'matplotlib', 'numpy', 'ete3', 'pandas', 'scipy'],
      python_requires='>=3.6',
      data_files=[("", ["LICENSE"])]
      )



import logging
import os
import sys


def make_sure_path_exists(path):
    """Create directory if it does not exist.

    Parameters
    ----------
    path : str
        The path to the directory which should be created.

    Returns
    -------
    bool
        True if the path exists.

    Raises
    ------
    OSError
        If an error was encountered while creating the directory.
    """
    if not path:
        # lack of a path qualifier is acceptable as this
        # simply specifies the current directory
        return True
    elif os.path.isdir(path):
        return True

    try:
        os.makedirs(path)
        return True
    except OSError:
        logger = logging.getLogger('timestamp')
        logger.error('Specified path could not be created: ' + path)
        raise


def is_executable(fpath):
    """Check if file is executable.

    This is a Python implementation of the linux
    command 'which'.

    Parameters
    ----------
    fpath : str
        Path to file.

    Returns
    -------
    boolean
        True if executable, else False.
    """

    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    """Return path to program.

    This is a Python implementation of the linux
    command 'which'.

    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    Parameters
    ----------
    program : str
        Name of executable for program.

    Returns
    -------
    str
        Path to executable, or None if it isn't on the path.
    """

    fpath, _fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file

    return None


def check_on_path(program, exit_on_fail=True):
    """Check if program is on the system path.

    Parameters
    ----------
    program : str
        Name of executable for program.
    exit_on_fail : boolean
        Exit program with error code -1 if program in not on path.

    Returns
    -------
    boolean
        True if program is on path, else False.
    """

    if which(program):
        return True

    if exit_on_fail:
        print('%s is not on the system path.' % program)
        sys.exit(1)

    return False


def rgba_hex(r, g, b, a=1):
    r, g, b = [int(x*a) for x in (r, g, b)]
    return '#%02x%02x%02x' % (r, g, b)


def hex_rgb(v):
    h = v.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))



class MetaTreeException(Exception):
    """Base class for all MetaTree exceptions."""

    def __init__(self, message):
        Exception.__init__(self, message)


class MetaTreeExit(MetaTreeException):
    """Raised when a controlled exit is requested."""

    def __init__(self, message=''):
        super(MetaTreeExit, self).__init__(message)
